from collections.abc import Callable
from datetime import datetime

from satori import Api
from satori.model import (
    Channel,
    ChannelType,
    Guild,
    Login,
    Member,
    MessageObject,
    PageDequeResult,
    PageResult,
    Role,
    User,
)
from satori.server import Adapter, Request
from satori.server.route import (
    ApproveParam,
    ChannelListParam,
    ChannelMuteParam,
    ChannelParam,
    ChannelUpdateParam,
    FriendListParam,
    GuildGetParam,
    GuildListParam,
    GuildMemberGetParam,
    GuildMemberKickParam,
    GuildMemberMuteParam,
    GuildMemberRoleParam,
    GuildXXXListParam,
    MessageListParam,
    MessageOpParam,
    MessageParam,
    ReactionCreateParam,
    ReactionDeleteParam,
    UserChannelCreateParam,
    UserGetParam,
)

from .message import OneBot11MessageEncoder, decode
from .utils import GROUP_AVATAR_URL, USER_AVATAR_URL, OneBotNetwork


def apply(adapter: Adapter, net_getter: Callable[[str], OneBotNetwork], login_getter: Callable[[str], Login]):
    @adapter.route(Api.LOGIN_GET)
    async def login_get(request: Request):
        return login_getter(request.self_id)

    @adapter.route(Api.MESSAGE_CREATE)
    async def message_create(request: Request[MessageParam]):
        net = net_getter(request.self_id)
        login = login_getter(request.self_id)
        encoder = OneBot11MessageEncoder(login, net, request.params["channel_id"])
        await encoder.send(request.params["content"])
        return encoder.results

    @adapter.route(Api.MESSAGE_GET)
    async def message_get(request: Request[MessageOpParam]):
        net = net_getter(request.self_id)
        result = await net.call_api("get_msg", {"message_id": int(request.params["message_id"])})
        if not result:
            raise RuntimeError(f"Failed to get message {request.params['message_id']}")
        sender: dict = result["sender"]
        user = User(str(sender["user_id"]), sender["nickname"], USER_AVATAR_URL.format(uin=sender["user_id"]))
        if result["message_type"] == "private":
            return MessageObject(
                str(result["message_id"]),
                await decode(result["message"], net),
                user=user,
                channel=Channel(f"private:{sender['user_id']}", ChannelType.DIRECT, sender["nickname"]),
            )
        member = Member(user, sender["nickname"], USER_AVATAR_URL.format(uin=sender["user_id"]))
        guild = Guild(request.params["channel_id"], avatar=GROUP_AVATAR_URL.format(group=request.params["channel_id"]))
        channel = Channel(request.params["channel_id"], ChannelType.TEXT)
        return MessageObject(
            str(result["message_id"]),
            await decode(result["message"], net),
            user=user,
            member=member,
            guild=guild,
            channel=channel,
        )

    @adapter.route(Api.MESSAGE_LIST)
    async def message_list(request: Request[MessageListParam]):
        net = net_getter(request.self_id)
        params = request.params
        direction = params.get("direction", "before")
        if direction != "before":
            raise RuntimeError("OneBot adapter only supports direction='before'")
        info = await net.call_api("get_version_info", {})
        app_name = info["app_name"]
        if app_name == "NapCat.Onebot":
            if params["channel_id"].startswith("private:"):
                messages = await net.call_api(
                    "get_friend_msg_history",
                    {
                        "user_id": params["channel_id"].removeprefix("private:"),
                        "message_seq": params.get("next", "0"),
                        "count": params.get("limit"),
                    },
                )
            else:
                messages = await net.call_api(
                    "get_group_msg_history",
                    {
                        "group_id": int(params["channel_id"]),
                        "message_seq": params.get("next", "0"),
                        "count": params.get("limit"),
                    },
                )
            next_ = str(len(messages) + int(params.get("next", "0")))
        elif app_name == "Lagrange.OneBot" and (mid := params.get("next")):
            if params["channel_id"].startswith("private:"):
                messages = await net.call_api(
                    "get_private_msg_history",
                    {
                        "user_id": params["channel_id"].removeprefix("private:"),
                        "message_id": int(mid),
                        "count": params.get("limit"),
                    },
                )
            else:
                messages = await net.call_api(
                    "get_group_msg_history",
                    {
                        "group_id": int(params["channel_id"]),
                        "message_id": int(mid),
                        "count": params.get("limit"),
                    },
                )
            next_ = messages[-1]["message_id"] if messages else mid
        elif app_name == "LLOneBot":
            if params["channel_id"].startswith("private:"):
                messages = await net.call_api(
                    "get_friend_msg_history",
                    {
                        "user_id": params["channel_id"].removeprefix("private:"),
                        "message_seq": params.get("next", "0"),
                        "count": params.get("limit"),
                    },
                )
            else:
                messages = await net.call_api(
                    "get_group_msg_history",
                    {
                        "group_id": int(params["channel_id"]),
                        "message_seq": params.get("next", "0"),
                        "count": params.get("limit"),
                    },
                )
            next_ = str(len(messages) + int(params.get("next", "0")))
        else:
            return PageDequeResult([])
        if not messages:
            return PageDequeResult([])
        dq = PageDequeResult([], next=next_)
        for result in messages:
            sender: dict = result["sender"]
            user = User(str(sender["user_id"]), sender["nickname"], USER_AVATAR_URL.format(uin=sender["user_id"]))
            if result["message_type"] == "private":
                dq.data.append(
                    MessageObject(
                        str(result["message_id"]),
                        await decode(result["message"], net),
                        user=user,
                        channel=Channel(f"private:{sender['user_id']}", ChannelType.DIRECT, sender["nickname"]),
                    )
                )
            else:
                member = Member(user, sender["nickname"], USER_AVATAR_URL.format(uin=sender["user_id"]))
                guild = Guild(
                    request.params["channel_id"], avatar=GROUP_AVATAR_URL.format(group=request.params["channel_id"])
                )
                channel = Channel(request.params["channel_id"], ChannelType.TEXT)
                dq.data.append(
                    MessageObject(
                        str(result["message_id"]),
                        await decode(result["message"], net),
                        user=user,
                        member=member,
                        guild=guild,
                        channel=channel,
                    )
                )
        if params.get("order") == "desc":
            dq.data.reverse()
        return dq

    @adapter.route(Api.MESSAGE_DELETE)
    async def message_delete(request: Request[MessageOpParam]):
        net = net_getter(request.self_id)
        await net.call_api("delete_msg", {"message_id": int(request.params["message_id"])})
        return

    @adapter.route(Api.CHANNEL_GET)
    async def channel_get(request: Request[ChannelParam]):
        net = net_getter(request.self_id)
        result = await net.call_api("get_group_info", {"group_id": int(request.params["channel_id"])})
        if not result:
            raise RuntimeError(f"Failed to get group {request.params['channel_id']}")
        return Channel(
            str(result["group_id"]),
            ChannelType.TEXT,
            result["group_name"],
            GROUP_AVATAR_URL.format(group=result["group_id"]),
        )

    @adapter.route(Api.GUILD_GET)
    async def guild_get(request: Request[GuildGetParam]):
        net = net_getter(request.self_id)
        result = await net.call_api("get_group_info", {"group_id": int(request.params["guild_id"])})
        if not result:
            raise RuntimeError(f"Failed to get group {request.params['guild_id']}")
        return Guild(str(result["group_id"]), result["group_name"], GROUP_AVATAR_URL.format(group=result["group_id"]))

    @adapter.route(Api.USER_CHANNEL_CREATE)
    async def user_channel_create(request: Request[UserChannelCreateParam]):
        user_id = str(request.params["user_id"])
        return Channel(f"private:{user_id}", ChannelType.DIRECT)

    @adapter.route(Api.CHANNEL_LIST)
    async def channel_list(request: Request[ChannelListParam]):
        net = net_getter(request.self_id)
        result: list[dict] = await net.call_api("get_group_list", {})  # type: ignore
        return PageResult(
            [
                Channel(
                    str(item["group_id"]),
                    ChannelType.TEXT,
                    item["group_name"],
                    GROUP_AVATAR_URL.format(group=item["group_id"]),
                )
                for item in result
            ]
        )

    @adapter.route(Api.GUILD_LIST)
    async def guild_list(request: Request[GuildListParam]):
        net = net_getter(request.self_id)
        result: list[dict] = await net.call_api("get_group_list", {})  # type: ignore
        return PageResult(
            [
                Guild(str(item["group_id"]), item["group_name"], GROUP_AVATAR_URL.format(group=item["group_id"]))
                for item in result
            ]
        )

    @adapter.route(Api.CHANNEL_UPDATE)
    async def channel_update(request: Request[ChannelUpdateParam]):
        net = net_getter(request.self_id)
        await net.call_api(
            "set_group_name",
            {"group_id": int(request.params["channel_id"]), "group_name": request.params["data"]["name"]},
        )
        return

    @adapter.route(Api.CHANNEL_MUTE)
    async def channel_mute(request: Request[ChannelMuteParam]):
        net = net_getter(request.self_id)
        await net.call_api(
            "set_group_whole_ban",
            {
                "group_id": int(request.params["channel_id"]),
                "enable": request.params["duration"] > 0,
            },
        )
        return

    @adapter.route(Api.GUILD_MEMBER_GET)
    async def guild_member_get(request: Request[GuildMemberGetParam]):
        net = net_getter(request.self_id)
        result = await net.call_api(
            "get_group_member_info",
            {"group_id": int(request.params["guild_id"]), "user_id": int(request.params["user_id"])},
        )
        if not result:
            raise RuntimeError(
                f"Failed to get member {request.params['user_id']} in group {request.params['guild_id']}"
            )
        user = User(str(result["user_id"]), result["nickname"], avatar=USER_AVATAR_URL.format(uin=result["user_id"]))
        return Member(user, result["card"], user.avatar, datetime.fromtimestamp(result["join_time"]))

    @adapter.route(Api.GUILD_MEMBER_LIST)
    async def guild_member_list(request: Request[GuildXXXListParam]):
        net = net_getter(request.self_id)
        result: list[dict] = await net.call_api(  # type: ignore
            "get_group_member_list",
            {"group_id": int(request.params["guild_id"])},
        )
        return PageResult(
            [
                Member(
                    User(
                        str(item["user_id"]),
                        item["nickname"],
                        avatar=USER_AVATAR_URL.format(uin=item["user_id"]),
                    ),
                    item["card"],
                    USER_AVATAR_URL.format(uin=item["user_id"]),
                    datetime.fromtimestamp(item["join_time"]),
                )
                for item in result
            ]
        )

    @adapter.route(Api.GUILD_MEMBER_KICK)
    async def guild_member_kick(request: Request[GuildMemberKickParam]):
        net = net_getter(request.self_id)
        await net.call_api(
            "set_group_kick",
            {
                "group_id": int(request.params["guild_id"]),
                "user_id": int(request.params["user_id"]),
                "reject_add_request": request.params.get("permanent", False),
            },
        )
        return

    @adapter.route(Api.GUILD_MEMBER_MUTE)
    async def guild_member_mute(request: Request[GuildMemberMuteParam]):
        net = net_getter(request.self_id)
        await net.call_api(
            "set_group_ban",
            {
                "group_id": int(request.params["guild_id"]),
                "user_id": int(request.params["user_id"]),
                "duration": int(request.params["duration"] / 1000),
            },
        )
        return

    @adapter.route(Api.GUILD_MEMBER_ROLE_SET)
    async def guild_member_role_set(request: Request[GuildMemberRoleParam]):
        net = net_getter(request.self_id)
        await net.call_api(
            "set_group_admin",
            {
                "group_id": int(request.params["guild_id"]),
                "user_id": int(request.params["user_id"]),
                "enable": request.params["role_id"] == "ADMINISTRATOR",
            },
        )
        return

    @adapter.route(Api.GUILD_MEMBER_ROLE_UNSET)
    async def guild_member_role_unset(request: Request[GuildMemberRoleParam]):
        net = net_getter(request.self_id)
        await net.call_api(
            "set_group_admin",
            {
                "group_id": int(request.params["guild_id"]),
                "user_id": int(request.params["user_id"]),
                "enable": request.params["role_id"] != "ADMINISTRATOR",
            },
        )
        return

    @adapter.route(Api.GUILD_ROLE_LIST)
    async def guild_role_list(request: Request[GuildXXXListParam]):
        return PageResult([Role("ADMINISTRATOR", "admin"), Role("MEMBER", "member"), Role("OWNER", "owner")])

    @adapter.route(Api.USER_GET)
    async def user_get(request: Request[UserGetParam]):
        net = net_getter(request.self_id)
        result = await net.call_api("get_stranger_info", {"user_id": int(request.params["user_id"])})
        if not result:
            raise RuntimeError(f"Failed to get user {request.params['user_id']}")
        return User(
            str(result["user_id"]),
            result["nickname"],
            result.get("remark"),
            USER_AVATAR_URL.format(uin=result["user_id"]),
        )

    @adapter.route(Api.FRIEND_LIST)
    async def friend_list(request: Request[FriendListParam]):
        net = net_getter(request.self_id)
        result: list[dict] = await net.call_api("get_friend_list", {})  # type: ignore
        return PageResult(
            [
                User(
                    str(item["user_id"]),
                    item["nickname"],
                    item.get("remark"),
                    USER_AVATAR_URL.format(uin=item["user_id"]),
                )
                for item in result
            ]
        )

    @adapter.route(Api.GUILD_APPROVE)
    async def guild_approve(request: Request[ApproveParam]):
        net = net_getter(request.self_id)
        await net.call_api(
            "set_group_add_request",
            {
                "flag": request.params["message_id"],
                "sub_type": "invite",
                "approve": request.params["approve"],
                "reason": request.params.get("comment", ""),
            },
        )
        return

    @adapter.route(Api.GUILD_MEMBER_APPROVE)
    async def guild_member_approve(request: Request[ApproveParam]):
        net = net_getter(request.self_id)
        await net.call_api(
            "set_group_add_request",
            {
                "flag": request.params["message_id"],
                "sub_type": "add",
                "approve": request.params["approve"],
                "reason": request.params.get("comment", ""),
            },
        )
        return

    @adapter.route(Api.FRIEND_APPROVE)
    async def friend_approve(request: Request[ApproveParam]):
        net = net_getter(request.self_id)
        await net.call_api(
            "set_friend_add_request",
            {
                "flag": request.params["message_id"],
                "approve": request.params["approve"],
            },
        )
        return

    @adapter.route(Api.REACTION_CREATE)
    async def reaction_create(request: Request[ReactionCreateParam]):
        net = net_getter(request.self_id)
        channel_id = request.params["channel_id"]
        info = await net.call_api("get_version_info", {})
        app_name = info["app_name"]
        if app_name == "LLOneBot":
            await net.call_api(
                "set_msg_emoji_like",
                {
                    "message_id": int(request.params["message_id"]),
                    "emoji": request.params["emoji"],
                },
            )
        elif app_name == "Lagrange.OneBot" and not channel_id.startswith("private:"):
            await net.call_api(
                "set_group_reaction",
                {
                    "group_id": int(channel_id),
                    "message_id": int(request.params["message_id"]),
                    "code": request.params["emoji"],
                    "is_add": True,
                },
            )
        elif app_name == "NapCat.Onebot":
            await net.call_api(
                "set_msg_emoji_like",
                {
                    "message_id": int(request.params["message_id"]),
                    "emoji_id": request.params["emoji"],
                    "set": True,
                },
            )
        elif app_name == "ws-plugin" and not channel_id.startswith("private:"):
            emj_type = 1 if int(request.params["emoji"]) < 5000 else 2
            await net.call_api(
                "set_reaction",
                {
                    "group_id": int(channel_id),
                    "message_seq": int(request.params["message_id"]),
                    "code": request.params["emoji"],
                    "is_add": True,
                    "type": emj_type,
                },
            )
        return

    @adapter.route(Api.REACTION_DELETE)
    async def reaction_delete(request: Request[ReactionDeleteParam]):
        net = net_getter(request.self_id)
        channel_id = request.params["channel_id"]
        info = await net.call_api("get_version_info", {})
        app_name = info["app_name"]
        if app_name == "LLOneBot":
            await net.call_api(
                "unset_msg_emoji_like",
                {
                    "message_id": int(request.params["message_id"]),
                    "emoji": request.params["emoji"],
                },
            )
        elif app_name == "Lagrange.OneBot" and not channel_id.startswith("private:"):
            await net.call_api(
                "set_group_reaction",
                {
                    "group_id": int(channel_id),
                    "message_id": int(request.params["message_id"]),
                    "code": request.params["emoji"],
                    "is_add": False,
                },
            )
        elif app_name == "NapCat.Onebot":
            await net.call_api(
                "set_msg_emoji_like",
                {
                    "message_id": int(request.params["message_id"]),
                    "emoji": request.params["emoji"],
                    "set": False,
                },
            )
        elif app_name == "ws-plugin" and not channel_id.startswith("private:"):
            emj_type = 1 if int(request.params["emoji"]) < 5000 else 2
            await net.call_api(
                "set_reaction",
                {
                    "group_id": int(channel_id),
                    "message_seq": int(request.params["message_id"]),
                    "code": request.params["emoji"],
                    "is_add": False,
                    "type": emj_type,
                },
            )
        return

    @adapter.route("*")
    async def internal_api(request: Request[dict]):
        net = net_getter(request.self_id)
        return await net.call_api(request.action.removeprefix("internal/"), request.params)
